; CROSSING. Written by Istvan 17 jul. 2007.

 LIST	p=16F84, R=DEC

 errorlevel 0,-305,-302
; Supressed Message[305]: Using default destination of 1 (f).
;	    Message[302]: Register in operand not in bank 0.
;			  Ensure that bank bits are correct.

 #include "P16F84.INC"	; Include header file

; Register File Usage (8 bit variables)

 CBLOCK	0x0C			;  Start variables after end of the uC regs

LCD_C				;  Lcd control byte
TempChar			;  Used When Sending Out Data to the LCD
State				;  Sw1 last state (bit 7 only)
AnimSpeed			;  Animation speed (smaller is faster)
AnimCount			;  Animation loop counter
Count				;  Unuversal counters
Count1
Delay1
Delay2

 ENDC

; Defines

#DEFINE	Offset		FSR	;  better to read
#DEFINE	YellowM 	0x4	;  PORTA,2
#DEFINE	SW1_Mode 	PORTB,7	;  Control Mode Switch
#DEFINE	RS		LCD_C,6	;  all LCD RS pin (1 for data, 0 for instr)
#DEFINE	SwState		State,7	;  Sw1 last state flag

#DEFINE	E1	0x10		;  RB4
#DEFINE	E2	0x20		;  RB5

#DEFINE	Red1		b'01010' ; PORTA bit patterns (traffic light contr.) 
#DEFINE	RedYellow1 	b'01100'
#DEFINE	Green1		b'10001'
#DEFINE	Yellow1		b'10100'

; Macros

LedSet	MACRO	param
	movlw	param
	movwf	PORTA
	ENDM

LcdNum	MACRO	param
	movlw	param
	movwf	LCD_C
	ENDM

; *************** CODE

	org	0		
	goto	Start		

	org	4		; Interrupt vector.
ever	goto	ever		; Sit in endless loop and do nothing.

; ***************  Tables  ***********

UserCharsDefTable		;  User chars (16 Bytes)
	addwf	PCL

;  Character user 1 (stop)

	retlw	b'01110'
	retlw	b'01110'
	retlw	b'00100'
	retlw	b'00100'
	retlw	b'00100'
	retlw	b'00100'
	retlw	b'01110'
	retlw	b'00000'

;  Character user 2 (walking)

	retlw	b'01110'
	retlw	b'01110'
	retlw	b'00100'
	retlw	b'00100'
	retlw	b'00100'
	retlw	b'01010'
	retlw	b'10001'
	retlw	b'00000'

;	retlw	0x00E		;  Character user 2 (other form)
;	retlw	0x00E
;	retlw	0x004
;	retlw	0x004
;	retlw	0x006
;	retlw	0x009
;	retlw	0x019
;	retlw	0x000

Messages
	addwf	PCL		;  PCL = $ + w

;	     "0123456789012345" ;  line lenght = 16 (LM016)

Mess_Stop			;  table to stop

     dt  1,0x7C, "            ",0x7C,1, 0	; 0x7C is a vertical line char.

Mess_ShiftOut			;  Output tables to shiftings

	dt    "               "	;  this is using with back shift only

Mess_Shift			;  sum lenght=22

	dt "   ",2,"   ",2,"   ",2,"   ",2,"   ",2,"  ", 0


; ***********  subroutines  (hw stack deep is max 8) ******

LcdReady
	call	GoHome		;  first pos in line 1
	call	SendUserCh2	;  User 2 = walking man (here: ready to go)

	movlw	0x80+15		;  last pos in line 1
	call	SendINS
	call	SendUserCh2

	movlw	0x80+0x40	;  first pos in line 2
	call	SendINS
	call	SendUserCh2

	movlw	0x80+0x40+15	;  last pos in line 2
	call	SendINS

SendUserCh2
	movlw	2

SendCHAR			;  Send the Character in W to the LCDs
	movwf	TempChar	;  temporaly save
	swapf	TempChar, w 	;  First send the high nybble
	bsf	RS		;  RS = 1 for data byte out
	call	NybbleOut	;  send the high nybble   	
	movf	TempChar, w	;  restore to send the low nybble

NybbleOut			;  Send a Nybble in W to the LCD
	andlw	0x0F		;  Low nybble only
	iorwf	LCD_C, w	;  Merged with Lcd controls (E and RS)
	movwf	PORTB		;  write out
	nop			;  delay
	andlw	b'11001111'	;  all "E" go to 0 but all others remain,
	movwf	PORTB		;   beacuse LCD "E" is working at falling edge!

Dlay32u				;  add a 32uS delay beacuse the min.
	movlw	( 32 / 4 )	;   instruction execution time is 37 uS.
	movwf	Delay1
	decfsz	Delay1
	 goto	$-1
	return

GoHome
	movlw	0x080		;  Go to the first pos of the 1st line

SendINS				;  Send an instruction to the LCD
	movwf	TempChar
	swapf	TempChar, w
	call	NybbleOutRS0
	movf	TempChar, w

NybbleOutRS0
	bcf	RS		;  RS = 0
	goto	NybbleOut

Dlay160				;  Delay 160 usecs
	movlw	( 160 / 4 )
	movwf	Delay1
	decfsz	Delay1
	 goto	$-1
	return

LongDlay
	clrw			;  Delay 256 * 5.9/8 = 192 mS
	goto	DlayW

Dlay5				;  Delay cca 5.9 msecs
	movlw	8
DlayW
	movwf	Delay1
	movwf	Delay2
	decfsz	Delay2
	 goto	$-1
	decfsz	Delay1
	 goto	$-3
	return

NewMode				;  return Z=1 if MOde Switch not changed
	movf	PORTB, w	;  get the whole portb to W
	andlw	0x80		;  only RB7 (the switch input)
	xorwf	State, w 	;  if equal to last state Z=1 else Z=0
	return

LcdStop
	call	GoHome
	movlw	Mess_Stop - Messages - 1
	movwf	Offset
	call	MessageOut

	movlw	Mess_Stop - Messages - 1
	movwf	Offset

MessageOut2
	movlw	0x80 + 0x40 	;  Go to the first pos of the 2nd line

MessageOutPosW
	call	SendINS

MessageOut
	movf	Offset, w
	call	Messages	;  get the next char
	iorlw	0		;  is it a zero end_sign?
	btfsc	STATUS, Z
	 return			;  yes, end
	call	SendCHAR
	incf	Offset
	goto	MessageOut

LcdShiftSlow
	movlw	200
	goto	LcdShift

LcdShiftFast
	movlw	100

LcdShift
	movwf	AnimSpeed
	call	GoHome
	movlw	Mess_Shift - Messages - 1
	movwf	Offset
	call	MessageOut
	movlw	10

ShiftFlash
	movwf	AnimCount

ShiftLoop0
	movlw	0x80 + 0x41 ;  line 2, pos at begin: first + 1 (bef. shift)
	movwf	Count1		;  store as cursor pos

	movlw	4		;  4 shift step
	movwf	Count

ShiftLoop1
	call	NewMode		; return Z=1 if MOde Switch not changed
	btfss	STATUS, Z	;  for faster mode switching  
	 goto	ShiftEnd	;  if changed, clear shifts and return

	movlw	Mess_Shift - Messages - 1 + 2	; initial shifting for line 2
	addwf	Count, w	;  modify stringtable pos to back shifting
	movwf	Offset
	movf	Count1, w	;  w = new cursor pos
	call	MessageOutPosW

	movlw	0x18 + 4  ;chars shift left (display as a window shifts right)
	call	SendINS

	movf	AnimSpeed, w
	call	DlayW

	incf	Count1		;  incr cursor pos
	decfsz	Count
	 goto	ShiftLoop1	;  next shift step

ShiftEnd
	movlw	0x2		;  clear shifts
	call	SendINS
	movlw	4		;  long instr., min 2 mSec needed!
	call	DlayW
	decfsz	AnimCount
	 goto	ShiftLoop0
	return

LcdShiftOut
	movlw	0x001		;  Clear the Display RAM
	call	SendINS
	call	Dlay5		;  can take up to 4.1 msecs

	movlw	Mess_Shift - Messages - 1 + 4
	movwf	Offset
	call	MessageOut

	movlw	0x80 + 0x41 ;  line 2, pos at begin: first + 1 (bef. shift)
	movwf	Count1		;  store as cursor pos

	movlw	1
	movwf	AnimCount

	movlw	14		;  14 shift step
	movwf	Count

ShiftOutLoop
	call	NewMode		; return Z=1 if MOde Switch not changed
	btfss	STATUS, Z	;  for faster mode switching  
	 goto	ShiftEnd	;  if changed, clear shifts and return

	movlw	Mess_ShiftOut - Messages - 1+3	; initial shifting for line 2
	addwf	Count, w	;  modify stringtable pos to back shifting
	movwf	Offset
	movf	Count1, w	;  w = new cursor pos
	call	MessageOutPosW

	movlw	0x18 + 4  ;chars shift left (display as a window shifts right)
	call	SendINS

	movf	AnimSpeed, w
	call	DlayW

	incf	Count1		;  incr cursor pos
	decfsz	Count
	 goto	ShiftOutLoop	;  next shift step
	goto	ShiftEnd

;*******************************************

Start

;16f84 HARDWARE INIC

	clrw			; Zero into W.
	movwf	PORTA		; Ensure PORTA is zero before we enable it.
	movwf	PORTB		; Ensure PORTB is zero before we enable it.
	bsf	STATUS,RP0	; Select Bank 1
	movwf	TRISA		; Set TRISA register, all output
	movlw	b'10000000'	; RB7 is an input, all other output
	movwf	TRISB		; Set TRISB register
	bcf	STATUS,RP0	; Reselect Bank 0.

; Ask SW1 at begin and save to SwState

	clrw
	movlw	State
	btfsc	SW1_Mode
	 bsf	SwState

	movlw	32
	call	DlayW		; Wait 20 msecs before LCD inic

; all LCD INIC

	LcdNum	E1+E2

	movlw	0x03		;  Reset Command
	call	NybbleOutRS0	;  Send the Nybble (instr)
	call	Dlay5		;  Wait 5 msecs before Sending again

	movlw	0x03
	call	NybbleOutRS0
	call	Dlay160		;  Wait 160 usecs before sending again

	movlw	0x03
	call	NybbleOutRS0
	call	Dlay160		;  Wait 160 usecs before sending the next instr.

	movlw	0x02		;  Set 4 Bit Mode
	call	NybbleOutRS0
	call	Dlay160

	movlw	0x028		;  It is a 2 Line Display
	call	SendINS

	movlw	0x008		;  Turn off the Display
	call	SendINS

	movlw	0x001		;  Clear the Display RAM
	call	SendINS
	call	Dlay5		;  can take up to 4.1 msecs

	movlw	0x006		;  Enable Cursor Move Direction
	call	SendINS

	movlw	0x00C		;  Turn the LCD Back On
	call	SendINS

; Defining two user chars

	movlw	0x048		;  Put the Cursor to the Start of CGRAM + 8
	call	SendINS		;   this is the address of the User char 1

	movlw	16		;  Send 2 Characters of 8 Bytes
	movwf	Count		;   for define the UserChars

	clrf	Offset
UserCharLoop
	movf	Offset, w	;  Get the next offset to Output
	incf	Offset
	call	UserCharsDefTable
	call	SendCHAR	;  Output the char
	decfsz	Count		;  Do 16x
	 goto	UserCharLoop

; **************** end of inics

	btfss	SwState		;  if flashing mode at begin, go to
	 goto	Flashing

InicNormal			;  needed at begin

	movlw	b'11000'	;  all red leds on beacuse
	movwf	PORTA		;  the first start or the switching back 
	LcdNum	E1+E2		;  from flashing mode is a special case 

	bsf	SwState		;  set last state as Normal mode

	movlw	100		;  walking people quickly "shift out"
	movwf	AnimSpeed
	call	LcdShiftOut	;  If SW1 changed, returns quickly

	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing

	call	LcdStop		;  inic LCDs with pictures as before red1
	LcdNum	E2
	call	LcdReady

	movlw	5		;  delay
	movwf	Count

InicLcdLoop
	call	LongDlay

	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing

	decfsz	Count
	 goto	InicLcdLoop

	goto	BeginNormMode
	 
NormLoop
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing

;Yellow1 (RedYellow2)

	LedSet	Yellow1

	LcdNum	E2
	call	LcdReady	;  first the fast (no loop) routine

	LcdNum	E1
	call	LcdShiftFast	;  If SW1 changed, returns quickly
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing

	call	LcdShiftOut
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing

;red1 (green2)

	LcdNum	E1
	call	LcdStop

BeginNormMode
	LedSet	Red1

	LcdNum	E2
	call	LcdShiftSlow
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing


;RedYellow1 ( yellow2)

	LedSet	RedYellow1

	LcdNum	E1
	call	LcdReady

	LcdNum	E2
	call	LcdShiftFast
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing

	call	LcdShiftOut
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	Flashing	;  if changed, go to yellow flashing


;Green1 (red2)

	LedSet	Green1

	LcdNum	E2
	call	LcdStop

	LcdNum	E1
	call	LcdShiftSlow

	goto	NormLoop

; ****************

Flashing

	bcf	SwState		;  set last state to flashing

	LcdNum	E1+E2		;  all LCD

	call	GoHome		;  line 1 picture for the shifting
	movlw	Mess_Shift - Messages - 1
	movwf	Offset
	call	MessageOut

	clrw			;  clear leds
	movwf	PORTA

	movlw	250		;  slow shifting
	movwf	AnimSpeed

FlashLoop
	call	NewMode		;  return Z=1 if MOde Switch not changed
	btfss	STATUS, Z
	 goto	InicNormal	;  if changed, go to normal mode inic

	movf	PORTA, w	;  toggle yellow leds
	xorlw	0x4
	movwf	PORTA

	movlw	1		;  AnimCount=1 for fast looping (flashing)
	call	ShiftFlash	;  If SW1 changed, returns quickly 

	goto	FlashLoop

 END
